# T_REREF -- Read reference features, shift, and stretch
# February 10, 2004
# By Doug Mink

# REREF reads a file produced by XCSAO containing the results of correlations
# between a template and a spectrum which is the same as the template but
# with distortions.  It then fits the distortions and modifies the image
# line positions in the template spectrum's database/id file so that the
# object spectrum can be reidentified reliably

include <pkg/gtools.h>
include <math/curfit.h>
include "identify.h"

define  MAXPTS          128

procedure t_reref

char	objfile[SZ_FNAME]	# Correlated file name
char	refin[SZ_FNAME]		# Input reference image
char	refout[SZ_FNAME]	# Output reference image
char	shiftfile[SZ_FNAME]	# XCSAO output file containing pixels and shifts
char	dbfile[SZ_FNAME]	# Path to id database file

bool	verbose, debug
int	i, nfeatures
int	ap[2]
double	dindef, dline0
char	s1[32], s2[32]
real	d1, d2, d3, d4
double	x[MAXPTS], y[MAXPTS], wts[MAXPTS], r[MAXPTS]
double	rmin			# Minimum R value for use of point in fit
int	ir

char	fitfunc[SZ_FNAME]	# Continuum fitting function type
char	logfile[SZ_FNAME]	# Log file
int	fitorder		# order of fit
int	niterate		# number of iterations for fit
real	grow			# radius around rejected points to replace
bool	interact		# Use interactive fitting (true or false)
int	npts
real	hirej, lowrej
real	smin
int	apout
int	fd, logfd
pointer	cv
pointer	ic
pointer	gopen(), gt_init()
int     fscan(), open()

int	clgeti()
real	clgetr()
bool	clgetb()
double	clgetd()
double	dcveval()
pointer	id		# ID pointer
pointer	gt, gp
int	strsearch()
errchk	id_dbread, id_map, id_dbsave

begin
	verbose = clgetb ("verbose")
	debug = clgetb ("debug")

	call clgstr ("object", objfile, SZ_FNAME)
	if (debug) {
	    call printf ("REREF: Crosscorrellated file is %s\n")
		call pargstr (objfile)
	    call flush (STDOUT)
	    }

	call strcpy ("database/rf", logfile, SZ_FNAME)
	call strcat (objfile, logfile, SZ_FNAME)
	if (debug) {
	    call printf ("REREF: Log file is %s\n")
		call pargstr (logfile)
	    call flush (STDOUT)
	    }
	iferr {logfd = open (logfile, NEW_FILE, TEXT_FILE)} then {
	    call delete (logfile)
	    logfd = open (logfile, NEW_FILE, TEXT_FILE)
	    }

        dindef = INDEFD
	call amovkd (0.d0, wts, MAXPTS)

	rmin = clgetd ("rmin")
	smin = clgetr ("smin")
	apout = clgetr ("apout")

#  Read X and shift vectors from XCSAO output
	call clgstr ("shiftfile", shiftfile, SZ_FNAME)
	fd = open (shiftfile, READ_ONLY, TEXT_FILE)
	if (debug) {
	    call printf ("REREF: Reading from shift file %s\n")
		call pargstr (shiftfile)
	    call flush (STDOUT)
	    verbose = TRUE
	    }
	npts = 0
	if (fd >0) {
	    while (fscan(fd) != EOF && npts < MAXPTS) {
		npts = npts + 1;
		call gargwrd (s1)
		call gargwrd (s2)
		call gargd (r[npts])
		call gargd (y[npts])
		call gargr (d1)
		call gargr (d2)
		call gargr (d3)
		call gargr (d4)
		call gargd (x[npts])
		if (r[npts] < rmin) {
		    wts[npts] = 0.0
		    }
		else {
		    wts[npts] = 1.0
		    }
		if (verbose) {
		    call printf ("%3d   %7.2f  %7.3f  %7.3f\n")
			call pargi (npts)
			call pargd (x[npts])
			call pargd (y[npts])
			call pargd (wts[npts])
		    }
		}
	    call close (fd)
	    if (verbose)
		call flush (STDOUT)
	    }

# Set the ICFIT pointer structure.
	call clgstr ("sfitfunc", fitfunc, SZ_FNAME)
	fitorder = clgeti ("sfitorder")
	niterate = 5
	hirej = smin
	lowrej = smin
	grow = 0
        call ic_open (ic)
        call ic_pstr (ic, "sample", "*")
        call ic_pstr (ic, "function", fitfunc)
        call ic_puti (ic, "naverage", 1)
        call ic_puti (ic, "order", fitorder)
        call ic_puti (ic, "niterate", niterate)
        call ic_putr (ic, "low", lowrej)
        call ic_putr (ic, "high", hirej)
        call ic_putr (ic, "grow", grow)
        call ic_pstr (ic, "ylabel", "")

#  Fit a function to X and Y (pixel and shift)
	if (verbose) {
	    call printf ("REREF: %d points to fit\n")
		call pargi (npts)
	    call flush (STDOUT)
	    }

#  If the interactive flag is set then use icg_fit to set the fitting
#  parameters.  Only done if task is run interactively as well.
	interact = clgetb ("sinteract")
	if (interact) {
	    iferr {
		gp = gopen ("stdgraph", NEW_FILE, STDGRAPH)
		gt = gt_init ()
		call gt_sets (gt, GTTYPE, "line")
		} then
		    call error (0, "Error opening `stdgraph'.")
	    call icg_fitd (ic,gp,"cursor",gt,cv,x,y,wts,npts)
	    call gt_free (gt)
	    call gclose (gp)
	    }

#  Do the fit non-interactively using ic_fit
	else {
	    call ic_fitd (ic,cv,x,y,wts,npts,YES,YES,YES,YES)
	    }

#  Read reference image name
	call clgstr ("refin", refin, SZ_FNAME)
	if (debug) {
	    call printf ("REREF: Input reference spectrum %s\n")
		call pargstr (refin)
	    call flush (STDOUT)
	    }
	if (refin[1] == EOS) {
	    call printf ("REREF: No reference image\n")
	    return
	    }

# Allocate the basic data structures.
	call id_init (id)
	if (debug) {
	    call printf ("REREF: ID structure initialized.\n")
	    call flush (STDOUT)
	    }
	call ic_open (ID_IC(id))
	if (debug) {
	    call printf ("REREF: ID structure opened.\n")
	    call flush (STDOUT)
	    }

# Open the image and return if there is an error.
	call strcpy (refin, ID_IMAGE(id), ID_LENSTRING)
	iferr (call id_map (id)) {
	    call error (0, "Cannot read reference image")
	    return
	    }
	iferr (call id_dbsave (id, ID_IMAGE(id))) {
#	    call printf ("REREF: Cannot read reference image line IDs\n")
#	    call error (0, "Cannot read reference image line IDs")
#	    return
	    }
	if (debug) {
	    call printf ("REREF: Image %s= %s opened in %s.\n")
		call pargstr (refin)
		call pargstr (ID_IMAGE(id))
		call pargstr (ID_DATABASE(id))
	    call flush (STDOUT)
	    }

# Get and save the reference database entry.
	call strcpy ("database", ID_DATABASE(id), ID_LENSTRING)
	call id_dbread (id, ID_IMAGE(id), ID_AP(id,1), NO, YES)
#	call printf ("REREF: Image database read.\n")
#	call flush (STDOUT)
	call id_saveap (id)
	if (debug) {
	    call printf ("REREF: Image database saved.\n")
	    call flush (STDOUT)
	    }

	nfeatures = ID_NFEATURES(id)
	if (nfeatures == 0)
	    call error (0, "No features in reference")
	else if (debug) {
	    call printf ("REREF: %d features in reference.\n")
		call pargi (nfeatures)
	    call flush (STDOUT)
	    }

# For each reference feature a new pixel value is computed using
# the fit function
	call amovkd (0.d0, wts, MAXPTS)
	do i = 1, nfeatures {
	    dline0 = PIX(id,i)
#	    if (debug) {
#		call printf ("%2d: %.3f\n")
#		    call pargi (i)
#		    call pargd (dline0)
#		call flush (STDOUT)
#		}

	    PIX(id,i) = dline0 + dcveval (cv, dline0)
	    x[i] = PIX(id,i)
	    y[i] = USER(id,i)
	    WTS(id,i) = 1.0
	    wts[i] = 1.0
	    if (verbose) {
		call printf ("%2d: %.3f => %.3f %.3f %.3f %.3f\n")
		    call pargi (i)
		    call pargd (dline0)
		    call pargd (PIX(id,i))
		    call pargd (FIT(id,i))
		    call pargd (USER(id,i))
		    call pargd (WTS(id,i))
		}
	    }

# Set the ICFIT pointer structure.
	call clgstr ("dfitfunc", fitfunc, SZ_FNAME)
	fitorder = clgeti ("dfitorder")
	niterate = 5
	hirej = 3.0
	lowrej = 3.0
	grow = 0
        call ic_open (ID_IC(id))
        call ic_pstr (ID_IC(id), "sample", "*")
        call ic_pstr (ID_IC(id), "function", fitfunc)
        call ic_puti (ID_IC(id), "naverage", 1)
        call ic_puti (ID_IC(id), "order", fitorder)
        call ic_puti (ID_IC(id), "niterate", niterate)
        call ic_putr (ID_IC(id), "low", lowrej)
        call ic_putr (ID_IC(id), "high", hirej)
        call ic_putr (ID_IC(id), "grow", grow)

#  Now fit a function to pixels (X) and wavelength (Y)
	if (verbose) {
	    call printf ("REREF: %d points to fit\n")
		call pargi (nfeatures)
	    do i = 1, nfeatures {
		call printf ("%3d   %7.2f  %7.3f  %5.3f\n")
		    call pargi (i)
		    call pargd (x[i])
		    call pargd (y[i])
		    call pargd (wts[i])
		}
	    call flush (STDOUT)
	    }

#  If the interactive flag is set then use icg_fit to set the fitting
#  parameters.  Only done if task is run interactively as well.
	interact = clgetb ("dinteract")
	if (interact) {
	    iferr {
		gp = gopen ("stdgraph", NEW_FILE, STDGRAPH)
		gt = gt_init ()
		call gt_sets (gt, GTTYPE, "line")
		} then
		    call error (0, "Error opening `stdgraph'.")
	    call icg_fitd (ic,gp,"cursor",gt,ID_CV(id),x,y,wts,nfeatures)
	    call gt_free (gt)
	    call gclose (gp)
	    }

#  Do the fit non-interactively using ic_fit
	else {
	    call ic_fitd (ID_IC(ID),ID_CV(id),x,y,wts,npts,YES,YES,YES,YES)
	    }

#  Recompute the fit wavelengths for the stretched and shifted pixels
	call id_fitdata (id)
	if (verbose) {
	    call printf ("\n i    Pix     Fit      User     Res\n")
	    }
	call fprintf (logfd, "\n i    Pix     Fit      User     Res\n")
	do i = 1, nfeatures {
	    call fprintf (logfd, "%2d: %8.3f %8.3f %8.3f %5.3f\n")
		call pargi (i)
		call pargd (PIX(id,i))
		call pargd (FIT(id,i))
		call pargd (USER(id,i))
		call pargd (USER(id,i)-FIT(id,i))
	    }
	call close (logfd)

	if (verbose) {
	    do i = 1, nfeatures {
		call printf ("%2d: %8.3f %8.3f %8.3f %5.3f\n")
		    call pargi (i)
		    call pargd (PIX(id,i))
		    call pargd (FIT(id,i))
		    call pargd (USER(id,i))
		    call pargd (USER(id,i)-FIT(id,i))
		}
	    call flush (STDOUT)
	    }
	if (debug) {
	    call printf ("REREF: Image %s= %s %s opened in %s.\n")
		call pargstr (refin)
		call pargstr (ID_IMAGE(id))
		call pargstr (ID_SECTION(id))
		call pargstr (ID_DATABASE(id))
	    call flush (STDOUT)
	    }
	call sprintf (ID_SECTION(id), ID_LENSTRING, "- Ap %d")
	    call pargi (apout)

# Write revised pixel numbers to the output database
	call clgstr ("refout", refout, SZ_FNAME)
	call strcpy ("database/id", dbfile, SZ_FNAME)
	call strcat (refout, dbfile, SZ_FNAME)
	if (debug) {
	    call printf ("REREF: Output reference spectrum %s.\n")
		call pargstr (refout)
	    call flush (STDOUT)
	    }
	call strcpy (refout, ID_IMAGE(id), ID_LENSTRING)
	ap[1] = Memi[ID_APS(id)]
	if (debug) {
	    call printf ("REREF: Aperture 1 from ID_APS is %d -> %d\n")
		call pargi (ap[1])
		call pargi (apout)
	    call flush (STDOUT)
	    }
	Memi[ID_APS(id)] = apout
	ap[1] = ID_AP(id,1)
	if (debug) {
	    call printf ("REREF: Aperture 1 from ID_AP is %d -> %d\n")
		call pargi (ap[1])
		call pargi (apout)
	    call flush (STDOUT)
	    }
	ap[1] = apout
	ID_AP(id,1) = apout
	ap[2] = ID_AP(id,2)
	if (debug) {
	    call printf ("REREF: Aperture 2 is %d\n")
		call pargi (ap[2])
	    call flush (STDOUT)
	    }
	ir = strsearch (refout, ".fits")
	if (ir > 0)
	    refout[ir-5] = EOS
	else {
	    ir = strsearch (refout, ".imh")
	    if (ir > 0)
		refout[ir-4] = EOS
	    }
	ID_NEWFEATURES(id) = YES
	ID_NEWCV(id) = YES
	ID_NEWDBENTRY(id) = YES
	call id_dbwrite (id, refout, ap, verbose)
	if (verbose) {
	    call show_id (id, "STDOUT")
	    }
	call show_id (id, logfile)

end


# SHOW_ID -- Show parameter information. (from ID_SHOW)

procedure show_id (id, file)

pointer	id			# ID pointer
char	file[ARB]		# File

char	line[SZ_LINE]
int	fd, i
double	pscoeff[16]
int	ncoeff, function

int	dcvstati()
int	open(), ic_geti()
errchk	open()

begin
	fd = open (file, APPEND, TEXT_FILE)

	call sysid (line, SZ_LINE)
	call fprintf (fd, "%s\n")
	call pargstr (line)
	call flush (fd)

	call fprintf (fd, "image %s\n")
	    call pargstr (ID_IMAGE(id))
	call fprintf (fd, "nsum %d\n")
	    call pargi (ID_NSUM(id,1))
	switch (ID_FTYPE(id)) {
	case EMISSION:
	    call fprintf (fd, "ftype emission\n")
	case ABSORPTION:
	    call fprintf (fd, "ftype absorption\n")
	}
	call flush (fd)
	switch (ID_LABELS(id)) {
	case 2:
	    call fprintf (fd, "labels index\n")
	case 3:
	    call fprintf (fd, "labels pixel\n")
	case 4:
	    call fprintf (fd, "labels coords\n")
	case 5:
	    call fprintf (fd, "labels user\n")
	case 6:
	    call fprintf (fd, "labels both\n")
	default:
	    call fprintf (fd, "labels none\n")
	}
	call flush (fd)
	call fprintf (fd, "maxfeatures %d\n")
	    call pargi (ID_MAXFEATURES(id))
	call flush (fd)
	call fprintf (fd, "match %g\n")
	    call pargr (ID_MATCH(id))
	call flush (fd)
	call fprintf (fd, "zwidth %g\n")
	    call pargr (ID_ZWIDTH(id))
	call flush (fd)
	call fprintf (fd, "fwidth %g\n")
	    call pargr (ID_FWIDTH(id))
	call flush (fd)
	call fprintf (fd, "database %s\n")
	    call pargstr (ID_DATABASE(id))
	call flush (fd)
	call fprintf (fd, "coordlist %s\n")
	    call pargstr (ID_COORDLIST(id))
	call flush (fd)
	call fprintf (fd, "cradius %g\n")
	    call pargr (ID_CRADIUS(id))
	call flush (fd)
	call fprintf (fd, "threshold %g\n")
	    call pargr (ID_THRESHOLD(id))
	call flush (fd)
	call fprintf (fd, "minsep %g\n")
	    call pargr (ID_MINSEP(id))
	call flush (fd)
	if (ID_CV(id) != NULL) {
	    call fprintf (fd, "function = %s\n")
	        call ic_gstr (ID_IC(id), "function", line, SZ_LINE)
		call pargstr (line)
	    call flush (fd)
	    call fprintf (fd, "order = %d\n")
		call pargi (ic_geti (ID_IC(id), "order"))
	    call flush (fd)

#  Convert coefficients to power series and print them
	    function = dcvstati (ID_CV(id), CVTYPE)
	    ncoeff = dcvstati (ID_CV(id), CVNCOEFF)
	    call fprintf (fd, "REREF: Function is %d, no. coeffs is %d\n")
		call pargi (function)
		call pargi (ncoeff)
	    call dcvpower (ID_CV(id), pscoeff, ncoeff)
	    call fprintf (fd, "Power Series Coefficients: ")
	    do i = 1, ncoeff {
	        call fprintf (fd, " %.6f")
		    call pargd (pscoeff[i])
		}
	    call fprintf (fd, "\n")

#	    call fprintf (fd, "Fit at first pixel = %0.8g\n")
#		call pargd (FITDATA(id,1))
#	    call flush (fd)
#	    call fprintf (fd, "Average fit interval = %0.8g\n")
#		call pargd ((FITDATA(id,ID_NPTS(id))-FITDATA(id,1))/
#		    (ID_NPTS(id)-1))
#	    call flush (fd)
	}

	call close (fd)
end

# Jan  7 2003	Add XCSAO output reading and fitting
# Jan 14 2003	Fit value is shift to be ADDED to template pixel 
# Apr 23 2003	Add fit of revised pixel values to wavelength
# Jun 30 2003	Write log file with residuals
# Jul  9 2003	Add rmin to drop points with low R values from fit
# Jul  9 2003	Add pwer series coefficients to database/rf file
# Oct 31 2003	Drop out after fitting input cross-correlation if no ref.spec.

# Feb  6 2004	Fix bug testing for reference file name null
# Feb 10 2004	Remove unswitched diagnostic printing
